/*

Copyright (c) 2006, Eric Tiedemann
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:


* Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

      * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

      * Neither the name of Eric Tiedemann nor the names of his
      contributors may be used to endorse or promote products derived
      from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <iostream>
#include <vector>

#include "bignum.h"
#include "rational.h"

using namespace e7;


#define display(expr) std::cerr << #expr << " = " << (expr) << std::endl;
#define display2(expr1, expr2) display(expr1); display(expr2);
#define display3(expr1, expr2, expr3) display(expr1); display(expr2); display(expr3);


#define NULLARY_METHOD(cmd, name) \
    } else if (strcmp(buf, cmd) == 0) {             \
      if (n < 1) {           \
        fprintf(stderr,            \
		"not enough values on stack for operation: %s\n",           \
		buf);           \
      } else {           \
        Bignum x = stack[n - 1];           \
	stack.pop_back();           \
        Bignum r = x.name();           \
        stack.push_back(r);           \
      }           \


#define BINARY_OP(cmd, op) \
    } else if (strcmp(buf, cmd) == 0) {             \
      if (n < 2) {           \
        fprintf(stderr,            \
		"not enough values on stack for operation: %s\n",           \
		buf);           \
      } else {           \
        Bignum a = stack[n - 2];           \
        Bignum b = stack[n - 1];           \
	stack.pop_back();           \
	stack.pop_back();           \
        Bignum r = a op b;           \
        stack.push_back(r);           \
      }           \


#define BINARY_PRED(cmd, op) \
    } else if (strcmp(buf, cmd) == 0) {             \
      if (n < 2) {           \
        fprintf(stderr,            \
		"not enough values on stack for operation: %s\n",           \
		buf);           \
      } else {           \
        Bignum a = stack[n - 2];           \
        Bignum b = stack[n - 1];           \
	stack.pop_back();           \
	stack.pop_back();           \
        Bignum r = Bignum(int64_t(a op b));           \
        stack.push_back(r);           \
      }           \


#define BINARY_METHOD(cmd, name) \
    } else if (strcmp(buf, cmd) == 0) {             \
      if (n < 2) {           \
        fprintf(stderr,            \
		"not enough values on stack for operation: %s\n",           \
		buf);           \
      } else {           \
        Bignum a = stack[n - 2];           \
        Bignum b = stack[n - 1];           \
	stack.pop_back();           \
	stack.pop_back();           \
        Bignum r = a.name(b);           \
        stack.push_back(r);           \
      }           \


#define BINARY_FUNC(cmd, func) \
    } else if (strcmp(buf, cmd) == 0) {             \
      if (n < 2) {           \
        fprintf(stderr,            \
		"not enough values on stack for operation: %s\n",           \
		buf);           \
      } else {           \
        Bignum a = stack[n - 2];           \
        Bignum b = stack[n - 1];           \
	stack.pop_back();           \
	stack.pop_back();           \
        Bignum r = func(a, b);           \
        stack.push_back(r);           \
      }           \


#define NULLARY_PRED(cmd, name) \
    } else if (strcmp(buf, cmd) == 0) {             \
      if (n < 1) {           \
        fprintf(stderr,            \
		"not enough values on stack for operation: %s\n",           \
		buf);           \
      } else {           \
        Bignum x = stack[n - 1];           \
	stack.pop_back();           \
        Bignum r = Bignum(int64_t(x.name()));           \
        stack.push_back(r);           \
      }           \



int main(int cargc, char ** argv) {

  e7::Bignum::init();

  std::vector<Bignum> stack;

  char buf[1024];

  while (fgets(buf, sizeof(buf), stdin) != NULL) {
    size_t n = stack.size();

    size_t nbuf = strlen(buf);
    assert(buf[nbuf - 1] == '\n');
    buf[nbuf - 1] = '\0';

    if (buf[0] == '\0') {
      ;				// no-op
    } else if (isdigit(buf[0])) {
      stack.push_back(Bignum(buf, 10, 0));
    } else if (buf[0] == '-' && isdigit(buf[1])) {
      stack.push_back(Bignum(buf + 1, 10, 1));
    } else if (strcmp(buf, "p") == 0) {
      for (size_t i = 0; i < n; i++) {
	Bignum q = stack[i];
	std::cout << q << std::endl;
	//std::cout << stack[i] << std::endl;
      }

    BINARY_FUNC("div", quotient);
    BINARY_FUNC("mod", remainder);
    BINARY_FUNC("expt", expt);
    BINARY_FUNC("gcd", gcd);
    BINARY_FUNC("lcm", lcm);

    BINARY_OP("+", +)
    BINARY_OP("-", -)
    BINARY_OP("*", *)

    BINARY_PRED("==", ==)
    BINARY_PRED("!=", !=)
    BINARY_PRED("<", <)
    BINARY_PRED("<=", <=)
    BINARY_PRED(">", >)
    BINARY_PRED(">=", >=)

    BINARY_OP("&", &)
    BINARY_OP("|", &)
    BINARY_OP("^", &)

    NULLARY_METHOD("abs", abs);
    NULLARY_METHOD("negative", negative);
    NULLARY_METHOD("complement", complement);
    NULLARY_METHOD("nbits", length_in_bits);
    NULLARY_METHOD("negative", negative);

    } else if (strcmp(buf, "bit-count") == 0) {             
      if (n < 1) {           
        fprintf(stderr,            
		"not enough values on stack for operation: %s\n",           
		buf);           
      } else {           
        Bignum x = stack[n - 1];           
	stack.pop_back();           
        Bignum r = x.length_in_bits();           
        stack.push_back(r);           
      }           

    } else if (strcmp(buf, "asl") == 0) {             
      if (n < 2) {           
        fprintf(stderr,            
		"not enough values on stack for operation: %s\n",           
		buf);           
      } else {           
        Bignum a = stack[n - 2];           
        int64_t b = stack[n - 1].to_int64();           
	stack.pop_back();           
	stack.pop_back();           
        Bignum r = a.asl(b);           
        stack.push_back(r);           
      }           


    NULLARY_PRED("oddp", oddp);
    NULLARY_PRED("evenp", evenp);
    NULLARY_PRED("zerop", zerop);
    NULLARY_PRED("positivep", positivep);
    NULLARY_PRED("negativep", negativep);

    } else if (strcmp(buf, "dup") == 0) {
      if (n < 1) {
	fprintf(stderr, 
		"not enough values on stack for operation: %s\n",
		buf);
      } else {
	stack.push_back(stack[n - 1]);
      }
    } else if (strcmp(buf, "swap") == 0) {
      if (n < 2) {
	fprintf(stderr, 
		"not enough values on stack for operation: %s\n",
		buf);
      } else {
	Bignum t = stack[n - 1];
	stack[n - 1] = stack[n - 2];
	stack[n - 2] = t;
      }
    } else if (strcmp(buf, "drop") == 0) {
      if (n < 1) {
	fprintf(stderr, 
		"not enough values on stack for operation: %s\n",
		buf);
      } else {
	stack.pop_back();
      }
    } else if (strcmp(buf, "q") == 0) {
      break;
    } else {
      fprintf(stderr, "unknown operation: %s\n", buf);
    }
  }

  if (ferror(stdin)) {
    perror("failure reading stdin");
  }

  return 0;
}
